#define XFree86Server
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/select.h>
#include <fcntl.h>
#include <pthread.h>
#include <X11/X.h>
#include <X11/Xatom.h>
#include <X11/Xlib.h>
#include <X11/Xlocale.h>
//#include <X11/Xos.h>
#include <X11/Xproto.h>
#include <X11/Xutil.h>
#define ATOM DWORD
#include <windows.h>
#include "winwindow.h"

#define WIN_CONNECT_RETRIES	5
#define WIN_CONNECT_DELAY	5
#define WIN_MSG_QUEUE_FNAME	"/dev/windows"
#define		WM_WM_X_EVENT	1

typedef struct _WMMsgNodeRec{
  winWMMessageRec msg;
  struct _WMMsgNodeRec* next;
}WMMsgNodeRec, *WMMsgNodePtr;

typedef struct _WMMsgQueueRec{
  struct _WMMsgNodeRec *pHead;
  struct _WMMsgNodeRec *pTail;
  pthread_mutex_t pmMutex;
  pthread_cond_t pcNotEmpty;
}WMMsgQueueRec, *WMMsgQueuePtr;

#if 0
typedef struct _ThreadControlRec {
  pthread_mutex_t pmMutex;
  pthread_cond_t pcCond;
  int nActive;
} ThreadControlRec;
#endif

HWND g_hwndMsg;
Display *g_pDisplay;
WMMsgQueueRec g_qMessage;
Atom g_atmWmProtos;
Atom g_atmWmDelete;

void
PushMessage (WMMsgNodePtr pNode)
{
  printf("PushMessage ()\n");
  pthread_mutex_lock (&g_qMessage.pmMutex);

  printf("PushMessage - mutex\n");
  pNode->next = NULL;

  if (g_qMessage.pTail != NULL)
    {
      //printf("\tg_qMessage.pTail->next = pNode\n");
      g_qMessage.pTail->next = pNode;
    }
  g_qMessage.pTail = pNode;

  if (g_qMessage.pHead == NULL)
    {
      //printf("\tg_qMessage.pHead = pNode;\n");
      g_qMessage.pHead = pNode;
    }
#if 0
  switch (pNode->msg.msg)
    {
    case WM_WM_MOVE:
      printf ("\tWM_WM_MOVE\n");
      break;
    case WM_WM_RAISE:
      printf ("\tWM_WM_RAISE\n");
      break;
    case WM_WM_LOWER:
      printf ("\tWM_WM_RAISE\n");
      break;
    case WM_WM_MAP:
      printf ("\tWM_WM_MAP\n");
      break;
    case WM_WM_UNMAP:
      printf ("\tWM_WM_UNMAP\n");
      break;
    case WM_WM_KILL:
      printf ("\tWM_WM_KILL\n");
      break;
    default:
      printf ("Unknown Message.\n");
      break;
    }
#endif
  printf("-PushMessage ()\n");
  pthread_mutex_unlock (&g_qMessage.pmMutex);
  pthread_cond_signal (&g_qMessage.pcNotEmpty);
}

WMMsgNodePtr
PopMessage ()
{
  WMMsgNodePtr pNode;
  printf ("PopMessage ()\n");
  pthread_mutex_lock (&g_qMessage.pmMutex);

  printf ("PopMessage - mutex\n");
  while (g_qMessage.pHead == NULL)
    {
      printf ("PopMessage - Waiting...\n");
      pthread_cond_wait (&g_qMessage.pcNotEmpty, &g_qMessage.pmMutex);
    }
  printf ("PopMessage - Cond signal.\n");
  
  pNode = g_qMessage.pHead;
  if (g_qMessage.pHead != NULL)
    {
      //printf("\tg_qMessage.pHead = g_qMessage.pHead->next;\n");
      g_qMessage.pHead = g_qMessage.pHead->next;
    }

  if (g_qMessage.pTail == pNode)
    {
      g_qMessage.pTail = NULL;
    }

  printf ("-PopMessage ()\n");
  pthread_mutex_unlock (&g_qMessage.pmMutex);
  return pNode;
}
	  
WMMsgNodePtr
PopSameMessage (int iWindow, int dwMsg)
{
  WMMsgNodePtr pNode;
  WMMsgNodePtr pPrevNode;
  WMMsgNodePtr pLastNode = NULL;
  
  pthread_mutex_lock (&g_qMessage.pmMutex);

  printf ("PopSameMessage ()\n");
  while (g_qMessage.pHead == NULL)
    {
      pthread_cond_wait (&g_qMessage.pcNotEmpty, &g_qMessage.pmMutex);
    }
  
  pPrevNode = NULL;
  
  for (pNode = g_qMessage.pHead; pNode != NULL; pNode = pNode->next)
    {
      if (pNode->msg.iWindow == iWindow && pNode->msg.msg == dwMsg)
	{
	  printf ("Remove ID %d\n", pNode->msg.dwID);
	  pLastNode = pNode;
	  if (pPrevNode == NULL)
	    {
	      g_qMessage.pHead = pNode->next;
	    }
	  else
	    {
	      pPrevNode->next = pNode->next;
	    }
	}
      else
	{
	  pPrevNode = pNode;
	}
    }

  if (g_qMessage.pTail == pLastNode)
    {
      g_qMessage.pTail = pPrevNode;
    }

  pthread_mutex_unlock (&g_qMessage.pmMutex);
  return pLastNode;
}

void
InitQueue()
{
  g_qMessage.pHead = NULL;
  g_qMessage.pTail = NULL;
  pthread_mutex_init (&g_qMessage.pmMutex, NULL);
  pthread_cond_init (&g_qMessage.pcNotEmpty, NULL);
}

void
GetWindowName (Display * pDpy, Window iWin, char **ppName)
{
  int nResult, nNum;
  char **ppList;
  XTextProperty xtpName;
  
  printf ("GetWindowName\n");
  *ppName = NULL;

  nResult = XGetWMName (pDpy, iWin, &xtpName);
  if (!nResult || !xtpName.value || !xtpName.nitems)
    {
      printf ("GetWindowName failed. No name\n");
      return;
    }

  if (xtpName.encoding == XA_STRING)
    {
      if (xtpName.value)
	{
	  *ppName = strdup ((char*)xtpName.value);
	  XFree (xtpName.value);
	}
    }
  else
    {
      XmbTextPropertyToTextList (pDpy, &xtpName, &ppList, &nNum);
      if (nNum && *ppList)
	{
	  XFree (xtpName.value);
	  *ppName = strdup (*ppList);
	  XFreeStringList (ppList);
	}
    }
  printf ("-GetWindowName\n");
}

int SendXMessage(Window w, Atom a, long x)
{
  XEvent e;

  e.type = ClientMessage;
  e.xclient.window = w;
  e.xclient.message_type = a;
  e.xclient.format = 32;
  e.xclient.data.l[0] = x;
  e.xclient.data.l[1] = CurrentTime;

  return XSendEvent(g_pDisplay, w, False, NoEventMask, &e);
}

void*
FlushWMMessageQueue (void* pArg)
{
  printf("FlushWMMessageQueue ()\n");
  for (;;)
    {
      WMMsgNodePtr pNode = PopMessage();
      WMMsgNodePtr pLastNode;
      
      if (pNode == NULL)
	{
	  printf ("Queue is Empty?\n");
	  exit (1);
	  //sleep (1);
	  //continue;
	}
      printf ("(%d ms)FlushWMMessageQueue MSG:%d ID:%d\n",
	      GetTickCount(), (int)pNode->msg.msg, (int)pNode->msg.dwID);

      //Window iWindow = (Window)GetProp(pWinPriv->hWnd, WIN_WID_PROP);
      if(1)
      switch (pNode->msg.msg)
	{
	case WM_WM_MOVE:
	  printf ("\tWM_WM_MOVE\n");
#if 0
	  pLastNode = PopSameMessage (pNode->msg.iWindow, pNode->msg.msg);
	  
	  if (pLastNode != NULL)
	    {
	      pNode = pLastNode;
	    }
#endif
	  printf ("(%d, %d)-(%d, %d)\n",
		   pNode->msg.iX, pNode->msg.iY,
		   pNode->msg.iX+pNode->msg.iWidth, pNode->msg.iY+pNode->msg.iHeight);
	  printf ("ID %d\n", pNode->msg.dwID);
	  
	  XMoveResizeWindow(g_pDisplay, pNode->msg.iWindow,
			    pNode->msg.iX, pNode->msg.iY,
			    pNode->msg.iWidth, pNode->msg.iHeight);
	  break;
	case WM_WM_RAISE:
	  printf ("\tWM_WM_RAISE\n");
	  XRaiseWindow(g_pDisplay, pNode->msg.iWindow);
	  break;
	case WM_WM_LOWER:
	  printf ("\tWM_WM_RAISE\n");
	  XLowerWindow(g_pDisplay, pNode->msg.iWindow);
	  break;
	case WM_WM_MAP:
	  printf ("\tWM_WM_MAP\n");
	  {
	    XWindowAttributes attr;
	    char *pszName;
	    
	    XMapWindow(g_pDisplay, pNode->msg.iWindow);
	    XGetWindowAttributes(g_pDisplay, pNode->msg.iWindow, &attr);
	    
	    XSetInputFocus(g_pDisplay, pNode->msg.iWindow, RevertToPointerRoot, CurrentTime);
	    //XSync (g_pDisplay, False);
	    if (!attr.override_redirect)
	      {
		DWORD dwExStyle;
		DWORD dwStyle;
		RECT rcNew;
		int iDx, iDy;
		
		GetWindowName(g_pDisplay, pNode->msg.iWindow, &pszName);
		SetWindowText (pNode->msg.hwndWindow, pszName);
		free (pszName);
		
		dwExStyle = GetWindowLongPtr (pNode->msg.hwndWindow,
					      GWL_EXSTYLE);
		dwStyle = GetWindowLongPtr (pNode->msg.hwndWindow,
					    GWL_STYLE);
		if (dwExStyle != WS_EX_APPWINDOW)
		  {
#if 1
		//pNode->msg.iX, pNode->msg.iY, pNode->msg.iWidth, pNode->msg.iHeight

		SetRect (&rcNew, pNode->msg.iX, pNode->msg.iY,
			pNode->msg.iX + pNode->msg.iWidth, pNode->msg.iY + pNode->msg.iHeight);

		    printf ("(%d, %d)-(%d, %d)\n",
			    rcNew.left, rcNew.top,
			    rcNew.right, rcNew.bottom);
		AdjustWindowRectEx (&rcNew, WS_POPUP|WS_SIZEBOX|WS_OVERLAPPEDWINDOW, FALSE, WS_EX_APPWINDOW);
		iDx = rcNew.left - pNode->msg.iX;
		iDy = rcNew.top - pNode->msg.iY;
		rcNew.left -= iDx;
		rcNew.right -= iDx;
		rcNew.top -= iDy;
		rcNew.bottom -= iDy;
#endif
		    printf ("(%d, %d)-(%d, %d)\n",
			    rcNew.left, rcNew.top,
			    rcNew.right, rcNew.bottom);
		    SetWindowLongPtr (pNode->msg.hwndWindow, GWL_EXSTYLE,
				      WS_EX_APPWINDOW);
		    SetWindowLongPtr (pNode->msg.hwndWindow, GWL_STYLE,
				      WS_POPUP|WS_SIZEBOX|WS_OVERLAPPEDWINDOW);
#if 1
		    SetWindowPos (pNode->msg.hwndWindow, HWND_TOP,
				  rcNew.left, rcNew.top,
				  rcNew.right - rcNew.left, rcNew.bottom - rcNew.top,
				  SWP_NOMOVE | /*SWP_NOZORDER | */SWP_FRAMECHANGED | SWP_SHOWWINDOW);
#else
		    SetWindowPos (pNode->msg.hwndWindow, NULL,
				  0, 0, 0, 0,
				  SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER
				  | SWP_FRAMECHANGED | SWP_SHOWWINDOW);
#endif
		  }
	      }
	  }
	  break;
	case WM_WM_UNMAP:
	  printf ("\tWM_WM_UNMAP\n");
	  XUnmapWindow(g_pDisplay, pNode->msg.iWindow);
	  break;
	case WM_WM_KILL:
	  printf ("\tWM_WM_KILL\n");
	  {
	    int i, n, found = 0;
	    Atom *protocols;

	    if (XGetWMProtocols(g_pDisplay, pNode->msg.iWindow, &protocols, &n))
	      {
		for (i=0; i<n; i++) if (protocols[i] == g_atmWmDelete) found++;
		XFree(protocols);
	      }
	    if (found)
	      SendXMessage(pNode->msg.iWindow, g_atmWmProtos, g_atmWmDelete);
	    else
	      XKillClient(g_pDisplay, pNode->msg.iWindow);
	  }
	  break;
	case WM_WM_X_EVENT:
	  FlushXEvents ();
	  break;
	default:
	  printf ("Unknown Message.\n");
	  //exit (1);
	  break;
	}
      free (pNode);
      XFlush (g_pDisplay);
    }
  printf("-FlushWMMessageQueue ()\n");
}

LRESULT CALLBACK
WndProc ( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  PCOPYDATASTRUCT pcdData;
  WMMsgNodePtr pNode;

  printf("WndProc\n");
  switch (uMsg)
    {
    case WM_CREATE:
      return 0;
    case WM_COPYDATA:
      printf("(%d ms)WM_COPYDATA\n", GetTickCount());
      printf("\t%X, %X\n", wParam, lParam);
      pcdData = (PCOPYDATASTRUCT)lParam;
      switch (pcdData->dwData)
	{
	case WMMSG_MSG:
	  pNode = (WMMsgNodePtr)malloc(sizeof(WMMsgNodeRec));
	  if (pNode == NULL)
	    {
	      printf("WM_COPYDATA - malloc fail.\n");
	      exit(1);
	    }
	  if (pcdData->cbData != sizeof(winWMMessageRec))
	    {
	      printf("WM_COPYDATA - pcdData->cbData invalid.\n");
	      break;
	    }
	  memcpy (&pNode->msg, pcdData->lpData, pcdData->cbData);
	  PushMessage (pNode);
	  break;
	default:
	  printf ("WM_COPYDATA - Unknown data type.\n");
	  break;
	}
      printf ("WM_COPYDATA - OK.\n");
      return 0;
    default:
      break;
    }
  return DefWindowProc(hwnd, uMsg, wParam, lParam); 
}

/*
 * Process any pending Windows messages
 */

void
FlushWindowsMessageQueue ()
{
  MSG			msg;
  
  //printf ("FlushWindowsMessageQueue ()\n");
  /* Flush the messaging window queue */
  while (PeekMessage (&msg, NULL, 0, 0, PM_REMOVE))
    {
      DispatchMessage (&msg);
    } 
  //printf ("-FlushWindowsMessageQueue ()\n");
}

Bool
FlushXEvents ()
{
#if 1
  XEvent		event;

  printf ("FlushXEvents ()\n");
  /* Process all pending events */
  while (XPending (g_pDisplay))
    {
      /* Get the next event - will not block because one is ready */
      XNextEvent (g_pDisplay, &event);

      /* Branch on the event type */
      switch (event.type)
	{
	}
    }
  printf ("-FlushXEvents ()\n");
#endif
  return True;
}
BOOL
InitializeApp()
{
	WNDCLASS wc;
	wc.style            = 0;
	wc.lpfnWndProc      = WndProc;
	wc.cbClsExtra       = 0;
	wc.cbWndExtra       = 0;
	wc.hInstance        = GetModuleHandle(NULL); 
	wc.hIcon            = NULL;
	wc.hCursor          = 0;
	wc.hbrBackground    = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName     = NULL;
	wc.lpszClassName    = WINDOW_CLASS_WM;
 
	if (!RegisterClass(&wc))
		return FALSE;

	g_hwndMsg = CreateWindow(WINDOW_CLASS_WM, WINDOW_TITLE_WM, WS_OVERLAPPED,	 //WS_POPUP,
			0,0,
			200, 200, NULL, NULL, GetModuleHandle(NULL), NULL);
	if (!g_hwndMsg)return FALSE;

	return TRUE;
}

int
ErrorHandler (Display *d, XErrorEvent *e)
{
  char pszErrorMsg[100];

  if (e->request_code == X_ChangeWindowAttributes
      && e->error_code == BadAccess)
    {
      printf ("There can be only ONE.\n");
      exit (1);
    }  


  XGetErrorText (d, e->error_code, pszErrorMsg, sizeof (pszErrorMsg));
  printf ("ERROR: %s\n", pszErrorMsg);

  if (e->error_code == BadWindow || e->error_code == BadMatch)
    {
      return 0;
    }
  
  exit(0);
  return 0;
}

int
main ()
{
  int                   iRetries = 0;
  int                   iReturn;
  int			iConnectionNumber;
  int			fdMessageQueue;
  fd_set		fdsRead;
  int			iMaxDescriptor;
  Bool			fReturn;
  struct timeval        tvTimeout;
  pthread_t             ptWMProc;
  pthread_t             ptXProc;
  MSG			msg;

  if (XInitThreads () == 0)
    {
      printf ("XInitThreads failed.\n");
      exit(1);
    }

  setvbuf (stdout, NULL, _IONBF, 0);

  setvbuf (stderr, NULL, _IONBF, 0);
  
  setlocale (LC_ALL, "");
  
  if (! InitializeApp()) return -1;

  do {
    g_pDisplay = XOpenDisplay (NULL);
    if (g_pDisplay == NULL)
      {
	++iRetries;
	sleep (WIN_CONNECT_DELAY);
	continue;
      }
    else
      break;
  } while (g_pDisplay == NULL && iRetries < WIN_CONNECT_RETRIES);

  if (g_pDisplay == NULL)
    {
      exit (1);
    }

  /* Get our connection number */
  iConnectionNumber = ConnectionNumber (g_pDisplay);

  /* Open a file descriptor for the windows message queue */
  fdMessageQueue = open (WIN_MSG_QUEUE_FNAME, O_RDONLY);
  if (fdMessageQueue == -1)
    {
      printf ("Failed opening %s\n", WIN_MSG_QUEUE_FNAME);
      exit (1);
    }

  /* Find max of our file descriptors */
  iMaxDescriptor = max (fdMessageQueue, iConnectionNumber) + 1;

  XSetErrorHandler (ErrorHandler);

  g_atmWmProtos = XInternAtom(g_pDisplay, "WM_PROTOCOLS", False);
  g_atmWmDelete = XInternAtom(g_pDisplay, "WM_DELETE_WINDOW", False);

  InitQueue ();
  
  if (pthread_create(&ptWMProc, NULL, FlushWMMessageQueue, NULL))
    {
      printf("main - pthread_create fail.\n");
      exit(1);
    }
#if 1
  while (GetMessage(&msg,NULL,0,0))
    {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
  return msg.wParam;
#else
  /* Loop for X events */
  while (1)
    {
      //FlushWMMessageQueue ();

      FD_ZERO (&fdsRead);
      FD_SET (fdMessageQueue, &fdsRead);
      //FD_SET (iConnectionNumber, &fdsRead);
      tvTimeout.tv_sec = 1;
      tvTimeout.tv_usec = 0;

      printf ("main - select events queue\n");
      /* Wait for a Windows event or an X event */
      iReturn = select (iMaxDescriptor,	/* Highest fds number */
			&fdsRead,	/* Read mask */
			NULL,		/* No write mask */
			NULL,		/* No exception mask */
			//&tvTimeout);	/* No timeout */
			NULL);	        /* No timeout */
      if (iReturn < 0)
	{
	  printf ("Call to select () failed: %d.  Bailing.\n", iReturn);
	  break;
	}
      
      /* Branch on which descriptor became active */
      if (FD_ISSET (fdMessageQueue, &fdsRead))
	{
	  /* Windows event ready */
#if 1
	  printf ("Windows event ready\n");
#endif
	  
	  /* Process Windows messages */
	  FlushWindowsMessageQueue ();
	}
      else if (FD_ISSET (iConnectionNumber, &fdsRead))
	{
#if 1
	  pNode = (WMMsgNodePtr)malloc(sizeof(WMMsgNodeRec));
	  if (pNode == NULL)
	    {
	      printf("malloc fail.\n");
	      exit(1);
	    }
	  pNode->msg.msg = WM_WM_X_EVENT;
	  PushMessage (pNode);
#else
	  /* X event ready */
#if 1
	  printf ("X event ready\n");
#endif

	  /* Process X events */
	  /* Exit when we see that server is shutting down */
	  //fReturn = FlushXEvents ();
	  if (!fReturn)
	    {
	      printf ("Caught WM_DELETE_WINDOW - shutting down\n");
	      break;
	    }
#endif
	}
#if 0
      else
	{
	  /* Unknown event ready */
	  printf ("Unknown event ready, exiting\n");
	  break;
	}
#endif
    }
#endif
  return 0;
}
